import importlib.metadata
from datetime import datetime

__version__ = importlib.metadata.version(__package__ or __name__)
__release_time__= datetime.strptime('2024-12-15T04:01:06','%Y-%m-%dT%H:%M:%S')
is_released_version=True



import cloudflare
from hiddifypanel.models import hconfig, ConfigEnum

__cf: cloudflare.Cloudflare   # type: ignore


def __prepare_cf_api_client() -> bool:
    '''Prepares cloudflare client if it's not already'''
    global __cf
    if __cf and isinstance(__cf, cloudflare.Cloudflare):
        return True

    if hconfig(ConfigEnum.cloudflare):
        __cf = cloudflare.Cloudflare(token=hconfig(ConfigEnum.cloudflare))
        if __cf and isinstance(__cf, cloudflare.Cloudflare):
            return True
    return False


def add_or_update_dns_record(domain: str, ip: str, dns_type: str = "A", proxied: bool = True) -> bool:
    '''This function cloud throw an exception'''
    if not __prepare_cf_api_client():
        return False

    zone_name = __extract_root_domain(domain)
    zone = __get_zone(zone_name)
    if zone:
        record = __get_dns_record(zone, domain)
        dns_name = domain[:-len(zone['name'])].replace('.', '')
        # if the input domain is root itself
        dns_name = '@' if not dns_name else dns_name
        data = {
            'name': dns_name,
            'type': dns_type, 'content': ip, 'proxied': proxied
        }
        if not record:
            api_res = __cf.zones.dns_records.post(zone['id'], data=data)
        else:
            api_res = __cf.zones.dns_records.put(zone['id'], record['id'], data=data)

        # validate api response
        if api_res['name'] == domain and api_res['type'] == dns_type and api_res['content'] == ip:
            return True
    return False


def delete_dns_record(domain: str) -> bool:
    '''Deletes a DNS record from cloudflare panel of user'''
    if not __prepare_cf_api_client():
        return False

    zone_name = __extract_root_domain(domain)
    zone = __get_zone(zone_name)
    record = __get_dns_record(zone, domain)
    if zone and record:
        api_res = __cf.zones.dns_records.delete(zone['id'], record['id'])
        if api_res['id'] == record['id']:
            return True
    return False


def __get_zone(zone_name: str) -> dict | None:
    zones = __cf.zones.get()
    for z in zones:
        if z['name'] == zone_name:
            return z
    return None


def __get_dns_record(zone, domain: str) -> dict | None:
    '''Returns dns record if exists'''
    dns_records = __cf.zones.dns_records(zone['id'])
    for r in dns_records:
        if r['name'] == domain:
            return r
    return None


def __extract_root_domain(domain: str) -> str:
    domain_parts = domain.split(".")
    if len(domain_parts) > 1:
        return ".".join(domain_parts[-2:])
    else:
        return domain



from enum import auto
import ipaddress
import re
from typing import Dict, List
from flask import request
from flask_babel import lazy_gettext as _
from sqlalchemy.orm import backref
from strenum import StrEnum
from sqlalchemy_serializer import SerializerMixin


from hiddifypanel.database import db
from hiddifypanel.models.config import hconfig
from .child import Child
from hiddifypanel.models.config_enum import ConfigEnum


class DomainType(StrEnum):
    direct = auto()
    sub_link_only = auto()
    cdn = auto()
    auto_cdn_ip = auto()
    relay = auto()
    reality = auto()
    old_xtls_direct = auto()
    worker = auto()
    fake = auto()

    # fake_cdn = "fake_cdn"
    # telegram_faketls = "telegram_faketls"
    # ss_faketls = "ss_faketls"


ShowDomain = db.Table('show_domain',
                      db.Column('domain_id', db.Integer, db.ForeignKey('domain.id'), primary_key=True),
                      db.Column('related_id', db.Integer, db.ForeignKey('domain.id'), primary_key=True)
                      )


class Domain(db.Model, SerializerMixin):
    id = db.Column(db.Integer, primary_key=True, autoincrement=True)
    child_id = db.Column(db.Integer, db.ForeignKey('child.id'), default=0)
    domain = db.Column(db.String(200), nullable=True, unique=False)
    alias = db.Column(db.String(200))
    sub_link_only = db.Column(db.Boolean, nullable=False, default=False)
    mode = db.Column(db.Enum(DomainType), nullable=False, default=DomainType.direct)
    cdn_ip = db.Column(db.Text(2000), nullable=True, default='')
    # port_index=db.Column(db.Integer, nullable=True, default=0)
    grpc = db.Column(db.Boolean, nullable=True, default=False)
    servernames = db.Column(db.String(1000), nullable=True, default='')
    # show_all=db.Column(db.Boolean, nullable=True)
    show_domains = db.relationship('Domain', secondary=ShowDomain,
                                   primaryjoin=id == ShowDomain.c.domain_id,
                                   secondaryjoin=id == ShowDomain.c.related_id,
                                   backref=backref('showed_by_domains', lazy='dynamic')
                                   )
    extra_params = db.Column(db.String(200), nullable=True, default='')

    def __repr__(self):
        return f'{self.domain}'

    def get_cdn_ips_parsed(self):
        ips = re.split('[ \t\r\n;,]+', self.cdn_ip.strip())
        res = set()
        for ip in ips:
            try:
                res.add(ipaddress.ip_address(ip))
            except:
                pass
        return res

    def to_dict(self, dump_ports=False, dump_child_id=False):
        data = {
            'domain': self.domain.lower(),
            'mode': self.mode,
            'alias': self.alias,
            'sub_link_only': self.sub_link_only,
            'child_unique_id': self.child.unique_id if self.child else '',  # type: ignore
            'cdn_ip': self.cdn_ip,
            'servernames': self.servernames,
            'grpc': self.grpc,
            'show_domains': [dd.domain for dd in self.show_domains],  # type: ignore
        }
        if dump_child_id:
            data['child_id'] = self.child_id
        if dump_ports:
            data["internal_port_hysteria2"] = self.internal_port_hysteria2
            data["internal_port_tuic"] = self.internal_port_tuic
            data["internal_port_reality"] = self.internal_port_reality
            data["need_valid_ssl"] = self.need_valid_ssl

        return data

    @staticmethod
    def from_schema(schema):
        return schema.dump(Domain())

    def to_schema(self):
        domain_dict = self.to_dict()
        from hiddifypanel.panel.commercial.restapi.v2.parent.schema import DomainSchema
        return DomainSchema().load(domain_dict)

    @property
    def need_valid_ssl(self):
        return self.mode in [DomainType.direct, DomainType.cdn, DomainType.worker, DomainType.relay, DomainType.auto_cdn_ip, DomainType.old_xtls_direct, DomainType.sub_link_only]

    @property
    def port_index(self):
        return self.id

    @property
    def internal_port_hysteria2(self):
        if self.mode not in [DomainType.direct, DomainType.relay, DomainType.fake]:
            return 0
        # TODO: check validity of the range of the port
        # print("child_id",self.child_id)
        return int(hconfig(ConfigEnum.hysteria_port, self.child_id)) + self.port_index

    @property
    def internal_port_tuic(self):
        if self.mode not in [DomainType.direct, DomainType.relay, DomainType.fake]:
            return 0
        # TODO: check validity of the range of the port
        return int(hconfig(ConfigEnum.tuic_port, self.child_id)) + self.port_index

    @property
    def internal_port_reality(self):
        if self.mode != DomainType.reality:
            return 0
        # TODO: check validity of the range of the port
        return int(hconfig(ConfigEnum.reality_port, self.child_id)) + self.port_index

    @classmethod
    def by_mode(cls, mode: DomainType) -> List['Domain']:
        domains = Domain.query.filter(Domain.mode == mode).all()
        if domains:
            return [d.domain for d in domains]
        return []

    @classmethod
    def modes_and_domains(cls) -> Dict[DomainType, List['Domain']]:
        return {mode: cls.by_mode(mode) for mode in DomainType}

    @classmethod
    def by_domain(cls, domain: str) -> 'Domain | None':
        return Domain.query.filter(Domain.domain == domain).first()

    @classmethod
    def get_panel_link(cls, child_id: int | None = None) -> str | None:
        if child_id is None:
            child_id = Child.current().id  # type: ignore
        domains = Domain.query.filter(Domain.mode.in_(
            [DomainType.direct, DomainType.cdn, DomainType.worker, DomainType.relay, DomainType.auto_cdn_ip, DomainType.old_xtls_direct, DomainType.sub_link_only]),
            Domain.child_id == child_id
        ).all()
        if not domains:
            return None
        return domains[0].domain

    @classmethod
    def get_domains(cls, always_add_ip=False, always_add_all_domains=False) -> List['Domain']:
        from hiddifypanel import hutils
        domains = []
        domains = Domain.query.filter(Domain.mode == DomainType.sub_link_only, Domain.child_id == Child.current().id).all()
        if not len(domains) or always_add_all_domains:
            domains = Domain.query.filter(Domain.mode.notin_([DomainType.fake, DomainType.reality])).all()

        if len(domains) == 0 and request:
            domains = [Domain(domain=request.host)]  # type: ignore
        if len(domains) == 0 or always_add_ip:
            domains += [Domain(domain=hutils.network.get_ip_str(4))]  # type: ignore
        return domains

    @classmethod
    def add_or_update(cls, commit=True, child_id=0, **domain):
        dbdomain = Domain.query.filter(Domain.domain == domain['domain']).first()
        if not dbdomain:
            dbdomain = Domain(domain=domain['domain'])  # type: ignore
            db.session.add(dbdomain)
        dbdomain.child_id = child_id

        dbdomain.mode = domain['mode']
        if (str(domain.get('sub_link_only', False)).lower() == 'true'):
            dbdomain.mode = DomainType.sub_link_only
        dbdomain.cdn_ip = domain.get('cdn_ip', '')
        dbdomain.alias = domain.get('alias', '')
        dbdomain.grpc = domain.get('grpc', False)
        dbdomain.servernames = domain.get('servernames', '')
        show_domains = domain.get('show_domains', [])
        dbdomain.show_domains = Domain.query.filter(Domain.domain.in_(show_domains)).all()
        if commit:
            db.session.commit()

    @classmethod
    def bulk_register(cls, domains, commit=True, remove=False, force_child_unique_id: str | None = None):
        from hiddifypanel.panel import hiddify
        child_ids = {}
        for domain in domains:
            child_id = hiddify.get_child(unique_id=force_child_unique_id)
            child_ids[child_id] = 1
            cls.add_or_update(commit=False, child_id=child_id, **domain)
        if remove and len(child_ids):
            dd = {d['domain']: 1 for d in domains}
            for d in Domain.query.filter(Domain.child_id.in_(child_ids)):
                if d.domain not in dd:
                    db.session.delete(d)

        if commit:
            db.session.commit()



from flask_babel import lazy_gettext as _
from sqlalchemy import func
from typing import Dict
import datetime

from hiddifypanel.drivers import user_driver
from hiddifypanel.models import *
from hiddifypanel.panel import hiddify
from hiddifypanel.database import db
from hiddifypanel import cache, hutils
from loguru import logger
to_gig_d = 1024**3

from celery import shared_task

@shared_task(ignore_result=False)
def update_local_usage():
    lock_key = "lock-update-local-usage"
    if not cache.redis_client.set(lock_key, "locked", nx=True, ex=600):
        return {"msg": "last update task is not finished yet."}
    try:
        res=update_local_usage_not_lock()
        cache.redis_client.set(lock_key, "locked", nx=False, ex=60)
        
        return res
    except Exception as e:
        cache.redis_client.set(lock_key, "locked", nx=False, ex=60)
        logger.exception("Exception in update usage")
        raise
        return {"msg": f"Exception in update usage: {e}"}

    # return {"status": 'success', "comments":res}
def update_local_usage_not_lock():
    
    try:
        res = user_driver.get_users_usage(reset=True)
        return _add_users_usage(res, child_id=0)
    except Exception as e:
        raise
        


def add_users_usage_uuid(uuids_bytes: Dict[str, Dict], child_id, sync=False):
    uuids_bytes = {u: v for u, v in uuids_bytes.items() if v}
    uuids = uuids_bytes.keys()
    users = User.query.filter(User.uuid.in_(uuids))
    dbusers_bytes = {u: uuids_bytes.get(u.uuid, 0) for u in users}
    _add_users_usage(dbusers_bytes, child_id, sync)  # type: ignore


def _reset_priodic_usage():
    last_usage_check: int = hconfig(ConfigEnum.last_priodic_usage_check) or 0
    import time
    current_time = int(time.time())
    if current_time - last_usage_check < 60 * 60 * 6:
        return
    # reset as soon as possible in the day
    if datetime.datetime.now().hour > 5 and current_time - last_usage_check < 60 * 60 * 24:
        return
    logger.debug("reseting user usage if needed")
    for user in User.query.filter(User.mode != UserMode.no_reset).all():
        if user.user_should_reset():
            logger.info(f"reseting user usage for {user.uuid}")
            user.reset_usage(commit=False)
    set_hconfig(ConfigEnum.last_priodic_usage_check, current_time)


def _add_users_usage(users_usage_data: Dict[User, Dict], child_id, sync=False):
    '''
    sync: when enabled, it means we have received usages from the parent panel
    '''
    res = {}
    have_change = False
    before_enabled_users = user_driver.get_enabled_users()

    daily_usage = {}
    today = datetime.date.today()
    changes = False
    for adm in AdminUser.query.all():
        daily_usage[adm.id] = DailyUsage.query.filter(DailyUsage.date == today, DailyUsage.admin_id == adm.id, DailyUsage.child_id == child_id).first()
        if daily_usage[adm.id] is None:
            logger.info(f"creating a new daily usage {today} admin={adm.id} child={child_id}")
            daily_usage[adm.id] = DailyUsage(date=today, admin_id=adm.id, child_id=child_id)
            db.session.add(daily_usage[adm.id])
            changes = True
        daily_usage[adm.id].online = User.query.filter(User.added_by == adm.id).filter(func.DATE(User.last_online) == today).count()
    if changes:
        db.session.commit()
    _reset_priodic_usage()

    # userDetails = {p.user_id: p for p in UserDetail.query.filter(UserDetail.child_id == child_id).all()}
    for user, uinfo in users_usage_data.items():
        usage_bytes = uinfo['usage']

        # UserDetails things
        # detail = UserDetail(user_id=user.id, child_id=child_id)
        # detail = userDetails.get(user.id)
        # if not detail:
        #     detail = UserDetail(user_id=user.id, child_id=child_id)
        #     db.session.add(detail)
        # if uinfo['devices'] != detail.connected_devices:
        #     detail.connected_devices = uinfo['devices']

        # Enable the user if isn't already
        if not before_enabled_users[user.uuid] and user.is_active:
            logger.info(f"Enabling disabled client {user.uuid} ")
            user_driver.add_client(user)
            send_bot_message(user)
            have_change = True

        # Check if there's new usage value
        if not isinstance(usage_bytes, int) or usage_bytes == 0:
            res[user.uuid] = "No usage"
        else:
            # Set new daily usage of the user
            if sync and daily_usage.get(user.added_by, daily_usage[1]).usage != usage_bytes:
                daily_usage.get(user.added_by, daily_usage[1]).usage = usage_bytes
            else:
                daily_usage.get(user.added_by, daily_usage[1]).usage += usage_bytes

            in_bytes = usage_bytes

            # Set new current usage of the user
            if sync and user.current_usage != in_bytes:
                user.current_usage = in_bytes
                # detail.current_usage_GB = in_gig
            else:
                user.current_usage += in_bytes
                # detail.current_usage = detail.current_usage or 0
                # detail.current_usage += in_bytes

            # Change last online time of the user
            user.last_online = datetime.datetime.now()
            # detail.last_online = datetime.datetime.now()

            # Set start date of user to the current datetime if it hasn't been set already
            if user.start_date is None:
                user.start_date = datetime.date.today()

            res[user.uuid] = f'{in_bytes/1000000:0.3f}MB'

        # Remove user from drivers(singbox, xray, wireguard etc.) if they're inactive
        # print(before_enabled_users[user.uuid], user.is_active)
        if before_enabled_users[user.uuid] and not user.is_active:
            logger.info(f"Removing enabled client {user.uuid} ")

            user_driver.remove_client(user)
            have_change = True
            res[user.uuid] = f"{res[user.uuid]} !OUT of USAGE! Client Removed"

    db.session.commit()  # type: ignore

    # Remove invalid users
    for uuid in before_enabled_users:
        if uuid in res:
            continue

        user = User.query.filter(User.uuid == uuid).first()
        if not user:
            user_driver.remove_client(User(uuid=uuid))
        elif not user.is_active:
            user_driver.remove_client(user)

    # print("------------------", res)
    # Apply the changes to the drivers
    if have_change:
        hiddify.quick_apply_users()

    # Sync the new data with the parent node if the data has not been set by the parent node itself and the current panel is a child panel
    if not sync and hutils.node.is_child():
        hutils.node.child.sync_users_usage_with_parent()

    return {"status": 'success', "comments": res, "date": hutils.convert.time_to_json(datetime.datetime.now())}


def send_bot_message(user):
    if not (hconfig(ConfigEnum.telegram_bot_token) or hutils.node.is_child()):
        return
    if not user.telegram_id:
        return
    from hiddifypanel.panel.commercial.telegrambot import bot, Usage
    try:
        msg = Usage.get_usage_msg(user.uuid)
        msg = _("User activated!") if user.is_active else _("Package ended!") + "\n" + msg
        bot.send_message(user.telegram_id, msg, reply_markup=Usage.user_keyboard(user.uuid))
    except BaseException:
        pass



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# NO CHECKED-IN PROTOBUF GENCODE
# source: hiddifypanel/panel/node/test.proto
# Protobuf Python Version: 5.27.2
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import runtime_version as _runtime_version
from google.protobuf import symbol_database as _symbol_database
from google.protobuf.internal import builder as _builder
_runtime_version.ValidateProtobufRuntimeVersion(
    _runtime_version.Domain.PUBLIC,
    5,
    27,
    2,
    '',
    'hiddifypanel/panel/node/test.proto'
)
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()




DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n\"hiddifypanel/panel/node/test.proto\"\x1b\n\x0cHelloRequest\x12\x0b\n\x03req\x18\x01 \x01(\t\"\x1c\n\rHelloResponse\x12\x0b\n\x03res\x18\x01 \x01(\t22\n\x05Hello\x12)\n\x08SayHello\x12\r.HelloRequest\x1a\x0e.HelloResponseb\x06proto3')

_globals = globals()
_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, _globals)
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'hiddifypanel.panel.node.test_pb2', _globals)
if not _descriptor._USE_C_DESCRIPTORS:
  DESCRIPTOR._loaded_options = None
  _globals['_HELLOREQUEST']._serialized_start=38
  _globals['_HELLOREQUEST']._serialized_end=65
  _globals['_HELLORESPONSE']._serialized_start=67
  _globals['_HELLORESPONSE']._serialized_end=95
  _globals['_HELLO']._serialized_start=97
  _globals['_HELLO']._serialized_end=147
# @@protoc_insertion_point(module_scope)



from flask import request, g
# from hiddifypanel.cache import cache



import datetime

from dotenv import dotenv_values
import os
import sys
from apiflask import APIFlask
from loguru import logger


from dynaconf import FlaskDynaconf

def create_app(*args, cli=False, **config):

    app = APIFlask(__name__, static_url_path="/<proxy_path>/static/", instance_relative_config=True, version='2.2.0', title="Hiddify API",
                   openapi_blueprint_url_prefix="/<proxy_path>/api", docs_ui='elements', json_errors=False, enable_openapi=not cli)
    # app = Flask(__name__, static_url_path="/<proxy_path>/static/", instance_relative_config=True)
    # app.asgi_app = WsgiToAsgi(app)
    
    for c, v in dotenv_values(os.environ.get("HIDDIFY_CFG_PATH", 'app.cfg')).items():
        if v.isdecimal():
            v = int(v)
        else:
            v = True if v.lower() == "true" else (False if v.lower() == "false" else v)
        app.config[c] = v
    dyn=FlaskDynaconf(app,settings_files=[os.environ.get("HIDDIFY_CFG_PATH", 'app.cfg')])

    if cli:
        app.config['EXTENSIONS']=[
            # "hiddifypanel.cache:init_app",
            "hiddifypanel.database:init_app",
            "hiddifypanel.panel.hlogger:init_cli",
            "hiddifypanel.panel.cli:init_app",
            "hiddifypanel.celery:init_app",
        ]
    else:
        app.config['EXTENSIONS']=[
            # "hiddifypanel.cache:init_app",
            "hiddifypanel.database:init_app",
            "hiddifypanel.panel.hlogger:init_app",
            "hiddifypanel.base_setup:init_app",
            "hiddifypanel.panel.common:init_app",
            "hiddifypanel.panel.common_bp:init_app",
            "hiddifypanel.panel.admin:init_app",
            "hiddifypanel.panel.user:init_app",
            "hiddifypanel.panel.commercial:init_app",
            "hiddifypanel.panel.node:init_app",
            "hiddifypanel.celery:init_app",
        ]
    


    app.config.update(config)  # Override with passed config
    
    app.config.load_extensions("EXTENSIONS")
    return app


def create_app_wsgi(*args, **kwargs):
    # workaround for Flask issue
    # that doesn't allow **config
    # to be passed to create_app
    # https://github.com/pallets/flask/issues/4170
    cli = ("hiddifypanel" in sys.argv[0] ) or (sys.argv[1] in ["update-usage", "all-configs", "admin_links", "admin_path"])

    app = create_app(cli=cli)
    return app



def create_celery_app():
    #     # workaround for Flask issue
    #     # that doesn't allow **config
    #     # to be passed to create_app
    #     # https://github.com/pallets/flask/issues/4170
    # print(kwargs)
    app = create_app(cli=True)
    return app.extensions["celery"]



from apiflask import Schema, fields
# region info


class PanelInfoOutputSchema(Schema):
    version = fields.String(description="The panel version")
# endregion


class PongOutputSchema(Schema):
    msg = fields.String(description="Pong Response")
# endregion
